︠edde3c7f-55d7-4ed6-b55a-6dcd0441fd24i︠
%html
For the sake of comparison we first give formulas for Weierstrass curves in projective coordinates; scroll down to see the formulas for Edwards curves
︡37962967-7b17-4ca4-a857-78640af5b558︡{"done":true,"html":"For the sake of comparison we first give formulas for Weierstrass curves in projective coordinates; scroll down to see the formulas for Edwards curves"}
︠7cf3f211-e1c5-4591-8184-4609451b1461s︠
def add_projective(P,Q,A):
    """addition algorithm for projective points an elliptic curve in short Weierstrass form y^2=x^3+Ax+B -- handles all cases (including points at infinity)"""
    if P[2] == 0: return Q
    if Q[2] == 0: return P
    x1=P[0]; y1=P[1]; z1=P[2]; x2=Q[0]; y2=Q[1]; z2=Q[2]
    # We use formulas from http://hyperelliptic.org/EFD/g1p/auto-shortw-projective.html#addition-add-1998-cmo-2
    x1z2 = x1*z2; x2z1 = x2*z1; y1z2 = y1*z2; y2z1 = y2*z1
    if x1z2 != x2z1:
        # usual case: P and Q are distinct and not opposites
        z1z2 = z1*z2
        u = y2z1-y1z2; uu = u^2
        v = x2z1-x1z2; vv = v^2; vvv = v^3
        r = vv*x1z2
        s = uu*z1z2 - vvv - 2*r
        x3 = v*s
        y3 = u*(r-s)-vvv*y1z2
        z3 = vvv*z1z2
    else:
        if y1z2 == -y2z1: return (0,1,0)      # P = -Q (includes case where P=Q is 2-torsion)
        # P = Q so we are doubling
        xx = x1^2; yy = y1^2; zz = z1^2
        w = A*zz + 3*xx
        s = 2*y1*z1; ss = s^2; sss = s*ss
        r = y1*s; rr = r^2
        b = (x1+r)^2-xx-rr
        h = w^2-2*b
        x3 = h*s
        y3 = w*(b-h)-2*rr
        z3 = sss
    return (x3,y3,z3)

def negate(P):
    """negates the point P on an elliptic curve in short Weierstrass form"""
    if P[2] == 0: return (0,1,0)
    return (P[0],-P[1],P[2])
    
def dbl_projective(P,A):
    """doubles a projective non-2-torsion point on an elliptic curve in short Weierstrass form"""
    x1 = P[0]; y1 = P[1]; z1 = P[2]
    xx = x1^2; yy = y1^2; zz = z1^2
    w = A*zz + 3*xx
    s = 2*y1*z1; ss = s^2; sss = s*ss
    r = y1*s; rr = r^2
    b = (x1+r)^2-xx-rr
    h = w^2-2*b
    x3 = h*s
    y3 = w*(b-h)-2*rr
    z3 = sss
    return (x3,y3,z3)

def equal_projective(P,Q):
    """Test whether two projective points are equal or not"""
    return P[0]*Q[2] == Q[0]*P[2] and P[1]*Q[2] == Q[1]*P[2]
    
def random_point(A,B):
    """generates a random affine point on the elliptic curve y^2 = x^2 + Ax + B"""
    F=A.parent()
    x0=F.random_element(); t=(x0^3+A*x0+B)
    while not t.is_square():
        x0=F.random_element(); t=(x0^3+A*x0+B)
    return (x0,sqrt(t),1)
︡695d0ef1-bb8f-42bc-9a2b-a5a856d9d119︡{"done":true}︡
︠ab16c6c7-9f86-4e74-826f-16843f80c7a2s︠
p=random_prime(2^256,lbound=2^255)
F=GF(p)
A=F(3); B=F.random_element()
P0=random_point(A,B)
Q0=random_point(A,B)
print "starting tests..."
timeit('add_projective(P0,Q0,A)',number=50000)
timeit('add_projective(P0,P0,A)',number=50000)
timeit('dbl_projective(P0,A)',number=50000) 
︡a4e80742-5133-4bda-bbd6-7b8aee61f627︡{"stdout":"starting tests...\n"}︡{"stdout":"50000 loops, best of 3: 13.6 µs per loop"}︡{"stdout":"\n"}︡{"stdout":"50000 loops, best of 3: 22.7 µs per loop"}︡{"stdout":"\n"}︡{"stdout":"50000 loops, best of 3: 20.3 µs per loop"}︡{"stdout":"\n"}︡{"done":true}︡
︠4d84d0ca-5d9f-4381-ac34-0d570ffe3c11i︠
%html
Note that distinct triples of proejctive coordinates may define the same projective point.
To compare points for equality we need to use the equal_projective function above (which also works for Edwards curves)
︡539d302e-88de-444c-b111-c3f80b9855b4︡{"done":true,"html":"Note that distinct triples of proejctive coordinates may define the same projective point.\nTo compare points for equality we need to use the equal_projective function above (which also works for Edwards curves)"}
︠63b2d563-329f-4e96-b329-ef4cd9dde155s︠
R1=add_projective(P0,Q0,A)
R2=add_projective(Q0,P0,A)
print "R1 =", R1
print "R2 =", R2
print "R1 == R2:",R1 == R2
print "equal_projective(R1,R2):",equal_projective(R1,R2)
︡d350c745-3b46-457e-a1fa-39d2a0e517e1︡{"stdout":"R1 = (88417280862405198524498904230150093073040285267296490962688023695194294232737, 81803801345615446712151118453505899941011682403746951648224039427654521101855, 46633157038940499649062681694319626916323919252431273559123417012920789472386)\n"}︡{"stdout":"R2 = (18524196991610979948129892573767159495541479256700257005341500524559971803562, 25137676508400731760477678350411352627570082120249796319805484792099744934444, 60308320815075678823566115109597625652257845271565474408906107206833476563913)\n"}︡{"stdout":"R1 == R2: False\n"}︡{"stdout":"equal_projective(R1,R2): True\n"}︡{"done":true}︡
︠8215e6a0-310e-4edd-989b-0e14ae66c8d2s︠
def add_edwards(P,Q,d):
    """adds two projective points on an elliptic curve in Edwards form (assumes d non-square, so operation is complete)"""
    # uses unoptimized formula derived in class (could save 1M with optimization)
    x1=P[0]; y1=P[1]; z1=P[2]; x2=Q[0]; y2=Q[1]; z2=Q[2]
    x1y2 = x1*y2; x2y1 = x2*y1
    r = z1*z2; rr = r^2; s = x1y2+x2y1; t = d*x1y2*x2y1; u = y1*y2 - x1*x2
    v = rr+t; w = rr-t
    x3 = r*s*w
    y3 = r*u*v
    z3 = v*w
    return (x3,y3,z3)
    
def dbl_edwards(P):
    """doubles a projective point on an elliptic curve in Edwards form"""
    # uses optimized formula from Bernstein-Lange
    x1 = P[0]; y1 = P[1]; z1 = P[2]
    B=(x1+y1)^2; C=x1^2; D=y1^2; E = C+D;
    J = E-2*z1^2; x3 = (B-E)*J; y3 = E*(C-D); z3 = E*J
    return (x3,y3,z3)
    
def neg_edwards(P):
    return (-P[0],P[1],P[2])
    
def random_edwards_point(d):
    """generates a random projective point on the Edwards curve x^2 + y^2 = 1+d*x^2*y^2 (assumes d non-square)"""
    F=d.parent()
    x0=F.random_element(); t=(1-x0^2)/(1-d*x0^2)
    while not t.is_square():
        x0=F.random_element(); t=(1-x0^2)/(1-d*x0^2)
    return (x0,sqrt(t),1)
    
def on_edwards_curve (P,d):
    x0=P[0]; y0=P[1]; z0=P[2]
    return z0^2*(x0^2+y0^2) == z0^4 + d*x0^2*y0^2
    
def is_identity_edwards (P):
    if P[0] == 0 and P[1] == P[2]: return true
    else: return false 
︡6734c383-f5da-48b7-a6fc-b79564e91532︡{"done":true}︡
︠380a6df9-eb7b-47b7-b9f0-e257f145293cs︠
p = 17
F=GF(p)
d = F.multiplicative_generator()
print "Points on Edwards curve x^2 + y^2 = 1 + %dx^2y^2 over F_%d"%(d,p)
for a in F:
    for b in F:
        if on_edwards_curve([a,b,1],d): print [a,b]
︡622cd674-6284-43e6-be28-16ff0cf47a45︡{"stdout":"Points on Edwards curve x^2 + y^2 = 1 + 3x^2y^2 over F_17\n"}︡{"stdout":"[0, 1]\n[0, 16]\n[1, 0]\n[2, 5]\n[2, 12]\n[3, 4]\n[3, 13]\n[4, 3]\n[4, 14]\n[5, 2]\n[5, 15]\n[7, 7]\n[7, 10]\n[10, 7]\n[10, 10]\n[12, 2]\n[12, 15]\n[13, 3]\n[13, 14]\n[14, 4]\n[14, 13]\n[15, 5]\n[15, 12]\n[16, 0]\n"}︡{"done":true}︡
︠fd85897a-9840-4cb2-8256-11d7979d59f0s︠
p=random_prime(2^256,lbound=2^255)
F=GF(p)
d=F(2);
while d.is_square(): d+=1;
P0=random_edwards_point(d)
Q0=random_edwards_point(d)
print "starting tests..."
timeit('add_edwards(P0,Q0,d)',number=100000)
timeit('add_edwards(P0,P0,d)',number=100000)
timeit('dbl_edwards(P0)',number=100000)
︡11f4e4de-e97b-4a4c-8d50-d0ae09ec0416︡{"stdout":"starting tests...\n"}︡{"stdout":"100000 loops, best of 3: 9.46 µs per loop"}︡{"stdout":"\n"}︡{"stdout":"100000 loops, best of 3: 9.17 µs per loop"}︡{"stdout":"\n"}︡{"stdout":"100000 loops, best of 3: 6.4 µs per loop"}︡{"stdout":"\n"}︡{"done":true}︡
︠4adc9c64-7273-425b-bb1b-d12c231190f1︠









